#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include "jit.h"

void *malloc_exec(size_t size);
void free_exec(void *ptr);

int ExecJIT_M(ContextM *m, MJIT *jit, int *err)
{
   typedef void (*EXEC)();
   char r;
   EXEC x = (EXEC)jit->ndata;
   x();
   if (err)
      *err = m->err;
   return m->regs[0];
}

typedef unsigned int (*WRITEOP)(unsigned char **, unsigned char **);

unsigned int WriteOp_End(unsigned char **in, unsigned char **out)
{
   unsigned char *inptr = *in;
   unsigned char *outptr = *out;

   if (!outptr)
      goto skipwrite;

   /* mov -1, [ebp+err] */
   *outptr++ = 0xC7;
   *outptr++ = 0x45;
   *outptr++ = offsetof(ContextM, err);
   *(int *)outptr = (int)-1;
   outptr += sizeof(int); 

   /* pop ebp */
   *outptr++ = 0x5D;
   /* ret */
   *outptr++ = 0xC3;

skipwrite:
   *out = outptr;

   return 9;
}

unsigned int WriteOp_Add(unsigned char **in, unsigned char **out)
{
   unsigned char *inptr = *in;
   unsigned char *outptr = *out;

   if (!outptr)
      goto skipwrite;
  
   /* mov eax, [ebp+regs[1]] */
   *outptr++ = 0x8A;
   *outptr++ = 0x45;
   *outptr++ = offsetof(ContextM, regs[1]);
   /* add [ebp+1], eax */
   *outptr++ = 0x00;
   *outptr++ = 0x45;
   *outptr++ = offsetof(ContextM, regs[0]);

skipwrite:
   *out = outptr;

   return 6;
}

unsigned int WriteOp_Multiply(unsigned char **in, unsigned char **out)
{
   unsigned char *inptr = *in;
   unsigned char *outptr = *out;

   if (!outptr)
      goto skipwrite;

   /* mov eax, [ebp+regs[0]] */
   *outptr++ = 0x8A;	/* mov */
   *outptr++ = 0x45;	/* rm */
   *outptr++ = offsetof(ContextM, regs[0]);
   /* mov edx, [ebp+regs[1]] */
   *outptr++ = 0x8A;
   *outptr++ = 0x55;
   *outptr++ = offsetof(ContextM, regs[1]);
   /* imul edx */
   *outptr++ = 0xF6;
   *outptr++ = 0xEA;
   /* mov [ebp+regs[0]], eax */
   *outptr++ = 0x88;
   *outptr++ = 0x45;
   *outptr++ = offsetof(ContextM, regs[0]);


skipwrite:
   *out = outptr;

   return 11;
}

unsigned int WriteOp_Set(unsigned char **in, unsigned char **out)
{
   unsigned char *inptr = *in;
   unsigned char *outptr = *out;
   unsigned char reg;
   char val;

   reg = *inptr++;
   val = *inptr++;

   if (!outptr)
      goto skipwrite;

   *outptr++ = 0xC6;   /* mov */
   *outptr++ = 0x45;   /* ebp */
   *outptr++ = 
      offsetof(ContextM, regs[reg]);   /* reg */
   *outptr++ = (unsigned char)val;   /* value */

skipwrite:
   *in = inptr;
   *out = outptr;

   return 4;
}

WRITEOP WriteOps[] = 
{
   WriteOp_Set,
   WriteOp_Add,
   WriteOp_Multiply,
   WriteOp_End
};

void RunJIT_M(ContextM *m, MJIT *jit)
{
   unsigned char *cip;
   unsigned char *ncip = NULL;
   unsigned char op;
   size_t codesize = 0;
   WRITEOP opcall;

   /* do a sample browse of the machine code */
   cip = m->bytecode; 
   do
   {
      op = *cip++;
      opcall = WriteOps[op];
      codesize += opcall(&cip, &ncip);
   } while (op != OP_END);

   /* 1 byte for prologuee */
   codesize += 6;
   ncip = malloc_exec(codesize);
   /* 1 byte from end of function */
   jit->code_size = codesize;
   jit->ndata = ncip;
   *ncip++ = 0x55;   /* push ebp */
   *ncip++ = 0xBD;   /* MOV EBP, */
   *(long *)ncip = (long)m;   /* m */
   ncip += sizeof(long);
   cip = m->bytecode;
   do
   {
      op = *cip++;
      opcall = WriteOps[op];
      opcall(&cip, &ncip);
   } while (op != OP_END);
}

void FreeJIT_M(MJIT *jit)
{
   free(jit->ndata);
}


#if defined __WIN32__
#include <windows.h>

void *malloc_exec(size_t size)
{
   return VirtualAlloc(NULL, size, MEM_RESERVE | MEM_COMMIT,
PAGE_EXECUTE_READWRITE);
}

void free_exec(void *ptr)
{
   VirtualFree(ptr, 0, MEM_RELEASE);
}

#elif defined __linux__
#include <sys/mman.h>
void *malloc_exec(size_t size)
{
   void *ptr = malloc(size);
   if (ptr)
      mprotect(ptr, size, PROT_READ | PROT_WRITE | PROT_EXEC);
   return ptr;
}

#define free_exec(ptr)   free(ptr)
#endif

