#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include "jit.h"

void *malloc_exec(size_t size);
void free_exec(void *ptr);

int ExecJIT_M(ContextM *m, MJIT *jit, int *err)
{
   typedef void (*EXEC)();
   char r;
   EXEC x = (EXEC)jit->ndata;
   x();
   if (err)
      *err = m->err;
   return m->regs[0];
}

void RunJIT_M(ContextM *m, MJIT *jit)
{
   unsigned char *cip;
   unsigned char *ncip = NULL;
   unsigned char op;
   unsigned char reg;
   char val;
   size_t codesize = 0;

   /* do a sample browse of the machine code */
   cip = m->bytecode; 
   do
   {
      op = *cip++;
      switch (op)
      {
         case OP_SET:
            cip += 2;
            codesize += 4;
            break;
         case OP_MUL:
            codesize += 11;
            break;
         case OP_ADD:
            codesize += 6;
            break;
         case OP_END:
            codesize += 9;
            break;
      }
   } while (op != OP_END);

   /* 1 byte for prologuee */
   codesize += 6;
   ncip = malloc_exec(codesize);
   /* 1 byte from end of function */
   jit->code_size = codesize;
   jit->ndata = ncip;
   *ncip++ = 0x55;   /* push ebp */
   *ncip++ = 0xBD;   /* MOV EBP, */
   *(long *)ncip = (long)m;   /* m */
   ncip += sizeof(long);
   cip = m->bytecode;
   do
   {
      op = *cip++;
      switch (op)
      {
         case OP_ADD:
           /* mov eax, [ebp+regs[1]] */
           *ncip++ = 0x8A;
           *ncip++ = 0x45;
           *ncip++ = offsetof(ContextM, regs[1]);
           /* add [ebp+1], eax */
           *ncip++ = 0x00;
           *ncip++ = 0x45;
           *ncip++ = offsetof(ContextM, regs[0]);
           break;
         case OP_SET:
           reg = *cip++;
           val = *cip++;
           *ncip++ = 0xC6;   /* mov */
           *ncip++ = 0x45;   /* ebp */
           *ncip++ = 
              offsetof(ContextM, regs[reg]);   /* reg */
           *ncip++ = (unsigned char)val;   /* value */
           break;
         case OP_END:
           /* mov -1, [ebp+err] */
           *ncip++ = 0xC7;
           *ncip++ = 0x45;
           *ncip++ = offsetof(ContextM, err);
           *(int *)ncip = (int)-1;
           ncip += sizeof(int); 
           /* pop ebp */
           *ncip++ = 0x5D;
           /* ret */
           *ncip++ = 0xC3;
           break;
         case OP_MUL:
           /* mov eax, [ebp+regs[0]] */
           *ncip++ = 0x8A;	/* mov */
           *ncip++ = 0x45;	/* rm */
           *ncip++ = offsetof(ContextM, regs[0]);
           /* mov edx, [ebp+regs[1]] */
           *ncip++ = 0x8A;
           *ncip++ = 0x55;
           *ncip++ = offsetof(ContextM, regs[1]);
           /* imul edx */
           *ncip++ = 0xF6;
           *ncip++ = 0xEA;
           /* mov [ebp+regs[0]], eax */
           *ncip++ = 0x88;
           *ncip++ = 0x45;
           *ncip++ = offsetof(ContextM, regs[0]);
           break;
      }
   } while (op != OP_END);
}

void FreeJIT_M(MJIT *jit)
{
   free(jit->ndata);
}


#if defined __WIN32__
#include <windows.h>

void *malloc_exec(size_t size)
{
   return VirtualAlloc(NULL, size, MEM_RESERVE | MEM_COMMIT,
PAGE_EXECUTE_READWRITE);
}

void free_exec(void *ptr)
{
   VirtualFree(ptr, 0, MEM_RELEASE);
}

#elif defined __linux__
#include <sys/mman.h>
void *malloc_exec(size_t size)
{
   void *ptr = malloc(size);
   if (ptr)
      mprotect(ptr, size, PROT_READ | PROT_WRITE | PROT_EXEC);
   return ptr;
}

#define free_exec(ptr)   free(ptr)
#endif

